##############################################
package aims2server;
##############################################

use strict;
use DBI;
use DBD::Oracle;
use MIME::Base64;
use Authen::Krb5;

##############################################
use subs qw(new SetUser AddImage RemoveImage GetImageByName AddHost RemoveHost GetHostByName EnablePXE DisablePXE GetKickstartFile UpdateKickstartFile RemoveKickstartFile EnableLGCY DisableLGCY);
##############################################

# Server Classes.
use aims2server::db;
use aims2server::conf;

use aims2server::landb;
use aims2server::host;
use aims2server::image;
use aims2server::pxe;

1;

##############################################
sub new
##############################################
{
   my($class, %args) = @_;
   my($self) = bless { _CONF => undef, _DB => undef, _SERVER => undef, _USER => undef }, $class;
   $self->SetServer;
   $self->ConnectToDatabase;
   $self->SetConf; 
   # CMF access , no kerberos
   if(defined($self->{_CLIENT}) && defined($self->{_CONF}->{'CMF_HOSTS'}) && $self->{_CLIENT} =~ /^($self->{_CONF}->{'CMF_HOSTS'})$/) {
    # idiotic e-groups design does not allow service accounts to be members
    # therefore once Ivan Deloose (deloose) retires ... 
    # this will stop working ... 
    $self->{_USER}='deloose';
   } else {
    $self->SetUser( $args{'username'}, $args{'password'} );
   }
   $self;  
}

##############################################
sub SetConf
##############################################
{
   my($self) = @_;
   
   my $dbh = $self->{_DB};
   unless($dbh){
      die "Error: Database connection has dropped.\n";
   }
   
   my $conf = aims2conf->new($dbh);
   $self->{_CONF} = $conf->get;
   $self->{_CONF}->{_MAPPINGS} = $conf->get_owner_mappings;
   
   $self;
}

##############################################
sub ConnectToDatabase
##############################################
{
   my($self) = @_;
   
   my $dbconn = $aims2conf::db_conn_string ? $aims2conf::db_conn_string : die "Error: Database server undefined.\n";
   my $dbuser = $aims2conf::db_user ? $aims2conf::db_user : die "Error: Database user undefined.\n";
   my $dbpass = $aims2conf::db_pass ? $aims2conf::db_pass : die "Error: Database password undefined.\n";
   my %dbattrs = %aims2conf::db_conn_attrib ? %aims2conf::db_conn_attrib : die "Error: Database attributes undefined.\n";
   
   my $db = aims2db->new();
   my $dbh = $db->Connect($dbconn, $dbuser, $dbpass, %dbattrs);
   
   $self->{_DB} = $dbh;
   
   $self;
}

##############################################
sub SetServer
##############################################
{
   my($self) = @_;
   
   my $server = `hostname -f`;
   chomp($server);
   $self->{_SERVER} = $server;
   
   my $client = $ENV{'REMOTE_HOST'};
   $self->{_CLIENT} = $client;
   
   # FIXME: IP address.
   # FIXME: Data & Time.
   
   # Is the daemon there?
   unless(-e '/mnt/data2/var/aims2sync.RUNNING'){
      die "Error: Synchronisation subsystem is not running.\n";
   }
   
   # Is the database there?
   if(-e '/mnt/data2/var/aims2sync.DBDOWN'){
      die "Error: Database not accessible: Try again later.\n";
   }
   
   $self;
}


##############################################
sub SetUser
##############################################
{
   my($self, $username, $password) = @_;

   # Authenticate and set the user array.
       
   # Credentials must be presented.
   unless($username and $password) {
      die "No authentication credentials presented.\n";
   }

   # The password given to us by the client is base64 encoded... we hope.
   $password = decode_base64($password);
    
   # Kerberos configuration values.
   my $version = $self->{_CONF}->{'KRB_VERSION'} ? $self->{_CONF}->{'KRB_VERSION'} : die "Error: KRB_VERSION is undefined.\n";
   my $service = $self->{_CONF}->{'KRB_SERVICE'} ? $self->{_CONF}->{'KRB_SERVICE'} : die "Error: KRB_SERVICE is undefined.\n";
   my $kt_file = $self->{_CONF}->{'KRB_KEYTAB'} ? $self->{_CONF}->{'KRB_KEYTAB'} : die "Error: KRB_KT is undefined.\n";   
   
   # Make sure that the $version strings match, and seperate from password.
   if ($password =~ /^$version=(.*)/s) {  
       ($version, $password) = split($version."=", $password);
   } else {
      die "ERROR: You are using an out-of-date aims2 client. Please update.\n";
   }
   
   # Initiate and create a new Auth Context Object.
   Authen::Krb5::init_context();
   my $ac = new Authen::Krb5::AuthContext;
   unless($ac){
      die "Error: Cannot create a new Authentication Context.\n";
   }
   
   # Resolve a service principal.
   my $sp = Authen::Krb5::sname_to_principal($self->{_SERVER},$service,KRB5_NT_SRV_HST);
   unless($sp){
      die "Error: Cannot resolve to a service principal.\n";
   }

   # Resolve to the keytab file.
   my $kt = Authen::Krb5::kt_resolve("FILE:$kt_file");
   unless($kt) {
      die "Error: Keytab file not available.\n";
   }
   
   # Try to decrypt the ticket.
   my $t = Authen::Krb5::rd_req($ac, $password, $sp, $kt);
   unless($t) {
      die "Authentication Failure: ", Authen::Krb5::error(), ".\n";
   } 
   
   my $user = $t->enc_part2->client->data;
   unless($user){
      die "Error: Couldn't extract user from ticket.\n";
   }
   
   undef($ac);
   Authen::Krb5::free_context();
  
   $self->{_USER} = $user;
   
   $self;
     
}

##############################################
sub AddImage
##############################################
{
   my($self, %args) = @_;

   # Upload a new image to aims2.

   unless(%args){
      die "Error: Input cannot be null.\n";
   }
   
   # Name of the image.
   my $name = $args{'name'} ? $args{'name'} : die "Error: No image name was provided.\n";
   $name = $self->validate($name);
   $name =~ tr/[a-z]/[A-Z]/;
   
   # Short description of the image.
   my $desc = $args{'desc'} ? $args{'desc'} : die "Error: No image description was provided.\n";
   $desc = $self->validate($desc);
   $desc =~ tr/[a-z]/[A-Z]/;
   
   # Arch of the image.
   my $arch = $args{'arch'};
   unless($arch and grep { $arch eq $_ } qw(i386 i686 x86_64 aarch64)){
      die "Error: Bad values provided for arch.\n";
   }
   
   # Extra bits for an image.
   my $kopts = $args{'kopts'} ? $args{'kopts'} : undef;
   my $groups = $args{'egroups'} ? $args{'egroups'} : undef;
   my $uefi = $args{'uefi'} ? $args{'uefi'}: undef;
   
   # Where the media is.
   my $initrd = $args{'initrd'} ? $args{'initrd'} : undef;
   my $vmlinuz = $args{'vmlinuz'} ? $args{'vmlinuz'} : die "Error: No kernel provided.\n"; 
  
   my $pxeboot = aims2image->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $pxeboot->upload_permission;
      $pxeboot->add({ 
         NAME => $name, 
         ARCH => $arch, 
         DESC => $desc,
         KOPTS => $kopts,
         GROUPS => $groups,
         INITRD => $initrd,
         VMLINUZ => $vmlinuz,
         UEFI => $uefi,
      });
   
   "Image $name uploaded to aims.";
   
}

##############################################
sub RemoveImage
##############################################
{
   my($self, %args) = @_;
   
   # Remove (erase) pxeboot media from aims2.
   
   my $dbh = $self->{_DB};
   
   my $name = $args{'imagename'} ? $args{'imagename'} : die "Error: Cannot remove pxe media. NAME is undefined.\n";
   $name = $self->validate($name);
   $name =~ tr/[a-z]/[A-Z]/;
   
   my $image = aims2image->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   
   # Only the owner and linuxsoft can remove the image.
   my $pxeboot = $image->get_pxeboot_by_name($name);
   my $owner = $pxeboot->{$name}{OWNER};
   unless($owner){
      die "Error: $name does not exist.\n";
   }
   $image->permission($owner);
   $image->remove($name);

   "Image $name removed from aims2.";
}

##############################################
sub GetImageByName
##############################################
{
   my($self, %args) = @_;
   
   # Fetch pxeboot entires matching $name.
   
   my $name = $args{'imagename'} ? $args{'imagename'} : die "Error: Cannot search for pxeboot. NAME is undefined.\n";
   my $arch = $args{'arch'} ? $args{'arch'} : undef;
   my $uefi = $args{'uefi'} ? $args{'uefi'} : undef;
 
   $name = $self->validate($name);
   $name =~ tr/[a-z]/[A-Z]/;
   
   my $image = aims2image->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   my $pxeboot = $image->get_pxeboot_by_name($name,$arch,$uefi);

   $pxeboot;
}

##############################################
sub AddHost
##############################################
{
   my( $self, %args ) = @_;
   
   # Register a host with aims2.
   
   my $dbh = $self->{_DB};
   
   # Clean up input data.
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Cannot remove host. Hostname is undefined.\n";
   }
   $hostname = $self->validate($hostname);
  
   my $kickstart = $self->validate($args{'kickstart'}) ? $args{'kickstart'} : undef;
   my $kopts = $self->validate($args{'kopts'}) ? $args{'kopts'} : undef;
   
   my $device = $self->GetDevice($hostname);
   
   # Do not take name from LanDB yet.. we may be dealing with an Interface alias.
   #$hostname = $device->{'DeviceName'};
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   # Register the device in the aims2 database.
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $host->permission($device);
      $host->remove($hostname);
      $host->add($hostname, $device, $kickstart, $kopts);
      
   # Kickstart CLOB?
   if($args{'ksclob'}){
      $kickstart = $args{'ksclob'};
      $host->insert_ks_record($hostname, $kickstart, "Uploaded");
   }
      
   "Host $hostname registered with aims2.";
}

##############################################
sub RemoveHost
##############################################
{
   my($self, %args) = @_;
   
   # Remove (de-register) a device from aims2.
   
   my $dbh = $self->{_DB};
   
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Add host failed. No hostname specified.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   #my $device = $self->GetDevice($hostname);
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   my $reg = $host->get_device_by_name($hostname);
   unless(values %$reg){
      die "Error: $hostname is not registered with aims2.\n";
   }
      #$host->permission($device);
      $host->remove($hostname);

   "Host $hostname removed from aims2.";
}

##############################################
sub GetHostByName
##############################################
{   
   my($self, %args) = @_;
   
   # Return a struct (reference) containing host data.
   
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Cannot fetch host. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_DB});
   my($hosts) = $host->get_device_by_name($hostname);

   $hosts;

}

##############################################
sub GetKickstartFile
##############################################
{
   my($self, %args) = @_;
   
   # Return kickstart to user (if permission).
   
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Unable to get kickstart. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   my $device = $self->GetDevice($hostname);
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   my $reg = $host->get_device_by_name($hostname);
   unless(values %$reg){
      die "Error: $hostname is not registered with aims2.\n";
   }
   $host->permission($device);
   
   my $kickstart = $host->get_host_kickstart($hostname);
   
   # hostname ks(encoded) source username lastupdated
   $kickstart;
   
}

##############################################
sub RemoveKickstartFile
##############################################
{
   die "Sorry, this feature is currently disabled. Use `updateks` instead.\n";
}

##############################################
sub UpdateKickstartFile
##############################################
{
   my($self, %args) = @_;
   
   # Update the kickstart file.
   
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Unable to get kickstart. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;

   my $kickstart = $self->validate($args{'kickstart'}) ? $args{'kickstart'} : undef;
   my $ksclob = $args{'ksclob'} ? $args{'ksclob'} : undef;
   
   my $device = $self->GetDevice($hostname);
   
   my $host = aims2host->new ( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   my $reg = $host->get_device_by_name($hostname);
   unless(values %$reg){
      die "Error: $hostname is not registered with aims2.\n";
   }

   $host->permission($device);

   if($ksclob){
      $host->update_ks_record($hostname, $ksclob, "Uploaded");
   }else{
      $host->updateks($hostname, $kickstart);
   }
      
   "Kickstart updated for $hostname.";
}


##############################################
sub EnableLGCY
##############################################
{
   my($self, %args) = @_;
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Unable to enable. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;

   my $device = $self->GetDevice($hostname);
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $host->permission($device);
   my $reg = $host->get_device_by_name($hostname);
   unless(values %$reg){
      die "Error: host $hostname is not registered with aims.\n";
   }

   my $pxe = aims2pxe->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   $pxe->enablelgcy($hostname);
   # legacy is for BIOS only, not UEFI !
   $pxe->changetobios($hostname);
 
   "Host $hostname legacy boot enabled (bios)."; 
}

##############################################
sub DisableLGCY
##############################################
{

  my($self, %args) = @_;

  my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Unable to disable. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   my $device = $self->GetDevice($hostname);  
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $host->permission($device);
   my $reg = $host->get_device_by_name($hostname);

   unless(values %$reg){
      die "Error: host $hostname is not registered with aims.\n";
   }

   my $pxe = aims2pxe->new(  $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $pxe->disablelgcy($hostname);
   
   "Host $hostname legacy boot disabled.";
}

##############################################
sub EnablePXE
##############################################
{
   my($self, %args) = @_;
   
   # Enable the device with pxeboot target.
   my $warn ="";
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Unable to enable. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   my $name = $args{'imagename'};
   unless($name){
      die "Error: Unable to enable. No pxeboot target provided.\n";
   }
 
   my $forever = 0;
   my $noexpiry = $args{'noexpiry'};
   if (defined($noexpiry) && $noexpiry =~/^noexpiry$/) {
     $forever = 1;
   }
   
   my $boottype;
   
   my $type= $args{'type'};
   if (defined($type)) {
    if ($type =~/^bios$/ || $type =~/^pxe$/) {
      $boottype = 0;
    } elsif ($type =~/^uefi$/) {
      $boottype = 1; 
    } elsif ($type =~/^arm64$/) {
      $boottype = 2;
    } else {
      die "Error: allowed boot types are: pxe|bios|uefi|arm64, but: $type specified.\n";
    }
   } else {
     $type = "bios";
     $boottype = 0;
   }

   $name = $self->validate($name);
   $name =~ tr/[a-z]/[A-Z]/;
   
   my $image = aims2image->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
   my $pxeboot = $image->get_pxeboot_by_name($name);
   unless( $pxeboot->{$name} ){
      die "Error: pxeboot target $name does not exist.\n";
   }
   
   my $device = $self->GetDevice($hostname);
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $host->permission($device);
   my $reg = $host->get_device_by_name($hostname);
   unless(values %$reg){
      die "Error: host $hostname is not registered with aims.\n";
   }
   
   my $pxe = aims2pxe->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $pxe->enable($hostname, $name ,$forever, $boottype);
      # legacy is for BIOS only
      if ($boottype != 0) {
        $pxe->disablelgcy($hostname);
      }

    if ($boottype == 1 && ($pxeboot->{$name}{UEFI} != 1 || !defined($pxeboot->{$name}{UEFI}))) {
       $warn="Warning: Non UEFI target selected for UEFI boot";
    }

   "Host $hostname enabled for $name ($type). $warn"; 
}

##############################################
sub DisablePXE
##############################################
{
   my($self, %args) = @_;
   
   # Set device to localboot (ie, disable pxe boot)
   
   my $hostname = $args{'hostname'};
   unless($hostname){
      die "Error: Unable to get kickstart. No hostname provided.\n";
   }
   $hostname = $self->validate($hostname);
   $hostname =~ tr/[a-z]/[A-Z]/;
   
   my $device = $self->GetDevice($hostname);  
   
   my $host = aims2host->new( $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $host->permission($device);
   my $reg = $host->get_device_by_name($hostname);
   unless(values %$reg){
      die "Error: host $hostname is not registered with aims.\n";
   }
   
   my $pxe = aims2pxe->new(  $self->{_DB}, $self->{_USER}, $self->{_CONF} );
      $pxe->disable($hostname);
   
   "Host $hostname set to localboot.";
}

##############################################
sub GetDevice 
##############################################
{
   my($self, $hostname) = @_;

   #die "($hostname)";
   
   # Get device struct from the network database.
   my $landbep = $self->{_CONF}->{'LANDB_EP'} ? $self->{_CONF}->{'LANDB_EP'} : die "Error: LANDB_EP is undefined.\n";
   my $landbproxy = $self->{_CONF}->{'LANDB_PROXY'} ? $self->{_CONF}->{'LANDB_PROXY'} : die "Error: LANDB_PROXY is undefined.\n";
   my $landbschema = $self->{_CONF}->{'LANDB_SCHEMA'} ? $self->{_CONF}->{'LANDB_SCHEMA'} : die "Error: LANDB_SCHEMA is undefined.\n";
   my $landbuser = $self->{_CONF}->{'LANDB_USER'} ? $self->{_CONF}->{'LANDB_USER'} : die "Error: LANDB_USER is undefined.\n";
   my $landbpass = $self->{_CONF}->{'LANDB_PASS'} ? $self->{_CONF}->{'LANDB_PASS'} : die "Error: LANDB_PASS is undefined.\n";

   my $landb = aims2landb->new();
   $landb->build_token($landbep, $landbproxy, $landbschema, $landbuser, $landbpass);
   my $device = $landb->get_device($hostname);
   
   $device;
   
}

##############################################
sub validate
##############################################
{
   my( $self, $value ) = @_;
   
   # allow / for PATHS
   # allow * for wilcards
   # allow numbers
   # allow letters
   # allow _
   # allow -
   # allow :
   
   # 'clean up' input values
   #if(defined($value) && $value ne ""){   
   #   if($value =~ /([\w\-\\\/\*\:\-\=]+)$/){
   #      $value = $1; 
   #   } else {
   #     die "Error: Validation is a bit strict.\n";
   #   }
   #}
   
   return $value;
}
